#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./paasta_cleanup_stale_nodes.py [options]

Removes terminated Kubernetes nodes from the Kubernetes API.

Command line options:

- -v, --verbose: Verbose output
- -n, --dry-run: Only report what would have been deleted
"""
import argparse
import logging
import sys
from typing import List
from typing import Sequence
from typing import Tuple

import boto3
from boto3_type_annotations.ec2 import Client
from botocore.exceptions import ClientError
from kubernetes.client import V1DeleteOptions
from kubernetes.client import V1Node
from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import get_all_nodes
from paasta_tools.kubernetes_tools import KubeClient

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Remove terminated Kubernetes nodes")
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", default=False
    )
    parser.add_argument(
        "-n", "--dry-run", action="store_true", dest="dry_run", default=False
    )
    args = parser.parse_args()
    return args


def nodes_for_cleanup(ec2_client: Client, nodes: Sequence[V1Node]) -> List[V1Node]:
    not_ready = [
        node
        for node in nodes
        if not is_node_ready(node)
        and "node-role.kubernetes.io/control-plane" not in node.metadata.labels
    ]
    terminated = terminated_nodes(ec2_client, not_ready)
    return terminated


def terminated_nodes(ec2_client: Client, nodes: Sequence[V1Node]) -> List[V1Node]:
    instance_ids = [node.spec.provider_id.split("/")[-1] for node in nodes]
    # if there are any instances that don't exist in the query to describe_instance_status
    # then amazon won't return the results for any, so we have to query the instances
    # one by one
    statuses = [
        does_instance_exist(ec2_client, instance_id) for instance_id in instance_ids
    ]
    for node, status in zip(nodes, statuses):
        log.debug(f"{node.metadata.name} exists: {status}")

    return [node for node, status in zip(nodes, statuses) if not status]


def does_instance_exist(ec2_client: Client, instance_id: str):
    try:
        instance = ec2_client.describe_instance_status(InstanceIds=[instance_id])
        if instance["InstanceStatuses"]:
            status = instance["InstanceStatuses"][0]["InstanceState"]["Name"]
            # see possible values at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instance_status
            # 'pending'|'running'|'shutting-down'|'terminated'|'stopping'|'stopped'
            # it's unlikely that we'll ever be in this situation with a pending node - the common case is that
            # the node is either running and been marked as not ready, or it's being shutdown
            if status not in ("running", "pending"):
                return False
            return True

        log.debug(
            f"no instance status in response for {instance_id}; assuming to have been terminated"
        )
        return False
    except ClientError as e:
        if e.response["Error"]["Code"] == "InvalidInstanceID.NotFound":
            log.debug(
                f"instance {instance_id} not found; assuming to have been terminated"
            )
            return False
        else:
            log.error(f"error fetching instance status for {instance_id}")
            raise e
    return True


def terminate_nodes(
    client: KubeClient, nodes: List[str]
) -> Tuple[List[str], List[Tuple[str, Exception]]]:
    success = []
    errors = []
    for node in nodes:
        try:
            body = V1DeleteOptions()
            client.core.delete_node(node, body=body, propagation_policy="foreground")
        except ApiException as e:
            errors.append((node, e))
            continue
        success.append(node)
    return (success, errors)


def is_node_ready(node: V1Node) -> bool:
    for condition in node.status.conditions:
        if condition.type == "Ready":
            return condition.status == "True"
    log.error(
        f"no KubeletReady condition found for node {node.metadata.name}. Conditions {node.status.conditions}"
    )
    return True


def main() -> None:
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    dry_run = args.dry_run

    kube_client = KubeClient()
    all_nodes = get_all_nodes(kube_client)
    log.debug(f"found nodes in cluster {[node.metadata.name for node in all_nodes]}")

    # we depend on iam credentials existing on the host for this to run.
    # anywhere else, and you'll need to set credentials using environment variables
    # we also make the assumption that all nodes are in the same region here
    region = all_nodes[0].metadata.labels["failure-domain.beta.kubernetes.io/region"]
    ec2_client = boto3.client("ec2", region)

    filtered_nodes = nodes_for_cleanup(ec2_client, all_nodes)
    if logging.DEBUG >= logging.root.level:
        log.debug(
            f"nodes to be deleted: {[node.metadata.name for node in filtered_nodes]}"
        )

    if not dry_run:
        success, errors = terminate_nodes(
            kube_client, [node.metadata.name for node in filtered_nodes]
        )
    else:
        success, errors = [], []
        log.info("dry run mode detected: not deleting nodes")

    for node_name in success:
        log.info(f"successfully deleted node {node_name}")

    for node_name, exception in errors:
        log.error(f"error deleting node: {node_name}: {exception}")

    if errors:
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import datetime
import re
import sys
from typing import Optional

from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import PaastaColors
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeploymentsJsonV2Dict
from paasta_tools.utils import format_timestamp
from paasta_tools.utils import get_pipeline_deploy_group_configs
from paasta_tools.utils import list_services
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import optionally_load_system_paasta_config
from paasta_tools.utils import parse_timestamp


def add_subparser(subparsers: argparse._SubParsersAction) -> None:
    parser = subparsers.add_parser(
        "get-image-version",
        help="Returns the value to be used for an image version, which will be used in automated redeploys of the same service SHA. If no deploy groups are configured for automated redeploys, will return no output.",
    )
    parser.add_argument(
        "-f",
        "--force",
        help="Force a brand new image_version, regardless if the latest build was recent",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--max-age",
        help="max age in seconds (default %(default)s)",
        type=int,
        default=2592000,  # TODO: Get from paasta system config
    )
    arg_service = parser.add_argument(
        "-s",
        "--service",
        help="Name of the service which you want to get the image version for.",
        required=True,
    )
    arg_service.completer = lazy_choices_completer(list_services)  # type: ignore
    parser.add_argument(
        "-c",
        "--commit",
        help="Commit to be used with generated image version",
        required=True,
    )
    parser.add_argument(
        "-y",
        "--soa-dir",
        help="A directory from which soa-configs should be read",
        default=DEFAULT_SOA_DIR,
    )
    parser.set_defaults(command=paasta_get_image_version)


def check_enable_automated_redeploys(service: str, soa_dir: str) -> bool:
    deploy_steps = get_pipeline_deploy_group_configs(service, soa_dir)
    enabled_by_default = (
        optionally_load_system_paasta_config().get_enable_automated_redeploys_default()
    )
    return any(
        [
            step.get("enable_automated_redeploys", enabled_by_default)
            for step in deploy_steps
        ]
    )


def extract_timestamp(image_version: str) -> Optional[datetime.datetime]:
    ts_str = re.match(r"^(?P<ts_str>[0-9]{8}T[0-9]{6})", image_version).group()
    return parse_timestamp(ts_str)


def should_generate_new_image_version(old: str, new: str, max_age: int) -> bool:
    # TODO: Handle additional criteria
    # Extract dates & compare
    try:
        age_diff = (extract_timestamp(new) - extract_timestamp(old)).total_seconds()

        if age_diff < max_age:
            return False
        print(
            PaastaColors.yellow(
                f"Old image version was {age_diff}s old, generating new one"
            ),
            file=sys.stderr,
        )
    except Exception as e:
        print(
            PaastaColors.red(
                f"Error: hit an exception {e} checking image version, will create new version {new}"
            ),
            file=sys.stderr,
        )
    return True


def get_latest_image_version(deployments: DeploymentsJsonV2Dict, commit: str) -> str:
    image_version = None
    # Image versions start with sortable timestamp
    # We only care about deployments for this sha; otherwise we will generate a new image_version
    sorted_image_versions = sorted(
        [
            deployment.get("image_version")
            for deployment in deployments["deployments"].values()
            if deployment.get("image_version") and deployment.get("git_sha") == commit
        ],
        reverse=True,
    )
    if sorted_image_versions:
        image_version = sorted_image_versions[0]
    return image_version


def paasta_get_image_version(args: argparse.Namespace) -> int:
    service = args.service
    soa_dir = args.soa_dir

    validate_service_name(service, soa_dir)

    # Check if any deploy groups have set enable_automated_redeploys
    if not check_enable_automated_redeploys(service, soa_dir):
        print(
            PaastaColors.red(
                f"Automated redeploys not enabled for {service}, returning no image_version"
            ),
            file=sys.stderr,
        )
        return 0

    current_timestamp = datetime.datetime.now()

    # TODO: Handle additional identifiers, flavor, etc
    new_image_version = format_timestamp(current_timestamp)

    if args.force:
        # Force a new version immediately
        print(new_image_version)
        return 0

    # get latest image_version of any deploy group from deployments.json
    deployments = load_v2_deployments_json(service, soa_dir)
    latest_image_version = get_latest_image_version(
        deployments.config_dict, commit=args.commit
    )

    if should_generate_new_image_version(
        old=latest_image_version, new=new_image_version, max_age=args.max_age
    ):
        print(new_image_version)
    else:
        print(latest_image_version)

    return 0



#!/usr/bin/env python
import argparse
from pathlib import Path

from kubernetes.client import V1Node

from paasta_tools.kubernetes_tools import KUBE_CONFIG_PATH
from paasta_tools.kubernetes_tools import KUBE_CONFIG_USER_PATH
from paasta_tools.kubernetes_tools import KubeClient


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Karpenter Habitat Corruption Workaround"
    )
    parser.add_argument("-c", "--cluster", required=True)
    parser.add_argument(
        "-k",
        "--kubeconfig",
        default=KUBE_CONFIG_PATH
        if Path(KUBE_CONFIG_PATH).exists()
        else KUBE_CONFIG_USER_PATH,
    )
    parser.add_argument(
        "-t", "--context", default=None  # -c is taken, so lets use the last letter :p
    )
    parser.add_argument(
        "--for-real",
        action="store_true",
    )
    parsed_args = parser.parse_args()

    if not parsed_args.context:
        if parsed_args.kubeconfig == KUBE_CONFIG_USER_PATH:
            # in the user kubeconfig, context names are just the cluster names
            parsed_args.context = parsed_args.cluster
        else:
            print(
                f"NOTE: no context specified - will use the current context selected in {parsed_args.kubeconfig} "
                "(or the KUBECONTEXT environment variable if set)."
            )

    return parsed_args


def is_affected_node(node: V1Node) -> bool:
    try:
        int(node.metadata.labels["yelp.com/habitat"])
        return True
    except ValueError:
        return False


def get_desired_habitat(node: V1Node) -> str:
    zone = node.metadata.labels["topology.kubernetes.io/zone"].replace("-", "")
    ecosystem = node.metadata.labels["yelp.com/ecosystem"]

    return f"{zone}{ecosystem}"


def main():
    args = parse_args()
    client = KubeClient(config_file=args.kubeconfig, context=args.context)
    for node in client.core.list_node().items:
        if not is_affected_node(node):
            continue

        if args.for_real:
            client.core.patch_node(
                name=node.metadata.name,
                body={
                    "metadata": {
                        "labels": {
                            "yelp.com/habitat": get_desired_habitat(node),
                        },
                    }
                },
            )
        else:
            print(
                f"Would have edited {node.metadata.name} in pool={node.metadata.labels['yelp.com/pool']} to have habitat={get_desired_habitat(node)} (from {node.metadata.labels['yelp.com/habitat']})",
            )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from paasta_tools.kubernetes_tools import get_all_kubernetes_services_running_here
from paasta_tools.mesos_tools import MesosSlaveConnectionError
from paasta_tools.tron_tools import tron_jobs_running_here
from paasta_tools.utils import _log
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config


def broadcast_log_all_services_running_here(line: str, soa_dir=DEFAULT_SOA_DIR) -> None:
    """Log a line of text to paasta logs of all services running on this host.

    :param line: text to log
    """
    system_paasta_config = load_system_paasta_config()
    cluster = system_paasta_config.get_cluster()
    services = get_all_services_running_here(cluster, soa_dir)
    for service, instance, _ in services:
        _log(
            line=line,
            service=service,
            instance=instance,
            component="monitoring",
            cluster=cluster,
        )


def get_all_services_running_here(cluster, soa_dir):
    try:
        tron_services = tron_jobs_running_here()
    except MesosSlaveConnectionError:
        tron_services = []

    try:
        kubernetes_services = get_all_kubernetes_services_running_here()
    except Exception:
        kubernetes_services = []

    return tron_services + kubernetes_services


def main() -> None:
    broadcast_log_all_services_running_here(sys.stdin.read().strip())


if __name__ == "__main__":
    main()
